//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 2001-                                               
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  func.hpp
//! \brief Function wrapper classes CTypedFunc and CFunc.

//! A wrapper holding a function with given input and output types.

typedef void (*funcPtr)(void); // generic pointer-to-function type

class CCoord;

//! A wrapper holding a function with fixed input and output types.

class CTypedFunc {
 public:
    const string outtype;
    const string intypes;
    funcPtr f; // the polymorphic function pointer
    CTypedFunc( const char* _outtype, const char* _intypes, funcPtr _f )
        : outtype(_outtype), intypes(_intypes), f(_f) {};
};

//! Metadata for an untyped function.

class CFuncMetadata {
 public:
    string tag;
    int narg;
    string explanation;
    int precedence;
    CFuncMetadata() {}
    CFuncMetadata( string _tag, int _narg, string _explanation, int _precedence=0 )
        : tag(_tag), narg(_narg), explanation(_explanation), precedence(_precedence) {}
};


//! A wrapper holding function metadata and implementations for various input and output types.

class CFunc : public CFuncMetadata {
 public:
    CFunc( const CFuncMetadata& _m ) : CFuncMetadata( _m ) {}
    vector<const CTypedFunc*> tyfuList; //! List of available typed function instances.
    const CTypedFunc* find_exact_tyfu( string intypes ) const;
    const CTypedFunc* find_tyfu( string intypes, bool want_error ) const;
    class CCoord coord( const CCoord& co ) const;
    class CCoord coord( const CCoord& co1, const CCoord& co2 ) const;
    class CCoord coord( const CCoord& co1, const CCoord& co2, const CCoord& co3 ) const;
};
