//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-                                
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  reg.hpp
//! \brief The registry.

//! Register reference: type and number (if applicable).

class CIdf {
 public:
    char typ;
    int num;

    CIdf( char _typ, int _num=0 ) : typ(_typ), num(_num) {}
    CIdf( const string s );
};

//! Direct access to registers ($-variables).

namespace NReg {
    extern double lastresult;
}


//! The registry.

class SRegistry {
 private:
    SRegistry() {}
    SRegistry(SRegistry const&);  //! To prevent copying; don't implement.
    void operator=(SRegistry const&); //! To prevent copying; don't implement.

    map<string,PObj> Map;       //! unsorted hash, for expression evaluation

public:
    static SRegistry& getInstance()
        {   static SRegistry instance; //! Instantiated on first use.
            return instance; }

    PObj find( string key ) const;
    PObj find_or_fail( string key ) const;
    
    
    void display_functions() const;
    
    void register_scalar( string key, PObj val );
    bool undef( string key );
};

extern SRegistry* GReg;
