//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* (C) Joachim Wuttke 1990-, C++ rewrite 2001-                            *//
//* http://apps.jcns.fz-juelich.de/frida                                   *//
//**************************************************************************//

//! \file  frida2.cpp
//! \brief Main program: initialize, then run interactive endless loop.


//! \mainpage FRIDA: fast reliable interactive data analysis

//! \par Homepage:
//!           http://apps.jcns.fz-juelich.de/frida

//! Useful entry points in this source documentation:
//! - initializations, main loop:
//!     <a href="frida2_8cpp_source.html">frida2.cpp</a>
//! - commands:
//!     <a href="commands_8cpp_source.html">commands.cpp</a>
//! - online memory:
//!     <a href="namespaceNOlm.html">namespace NOlm</a>
//! - data model root:
//!     <a href="classCOlo.html">class COlo</a>
//!
//! Used libraries:
//! - Minor utilities:
//!     <a href="../frida-libs/trivia/index.html">libtrivia</a>
//! - Dialogue interface:
//!     <a href="../frida-libs/readplus/index.html">libreadplus</a>


#include <string>
#include "config_src.hpp"
#include "toplevel.hpp"

using namespace std;


//! Frida main program.

int main(int argc, char* argv[])
{
    // Initialize runtime environment.
    CFrida frida;

#ifdef CONFDIR
    // Execute initialization script.
    frida.execute_file( CONFDIR "/frida.ini" );
#endif

    // Execute scripts from command line (used in particular by shebang files).
    for( int iarg=1; iarg<argc; ++iarg )
        frida.execute_file( argv[iarg] );

    // Interactive main loop.
    frida.interactive();
}
